package App::This2that;

use 5.010001;
use strict;
use warnings;

use Alarm::Concurrent qw( setalarm );
use Carp;
use Data::Dump qw( pp );
use File::Basename;
use File::Copy qw( move );
use File::HomeDir;
use File::Path qw( make_path );
use File::Spec;
use Getopt::Long;
use IO::Dir;
use Linux::Inotify2;
use POSIX qw( strftime );
use Scalar::Util qw( looks_like_number );
use String::ShellQuote;
use YAML::Syck;

our $VERSION = '0.02';
$VERSION = eval $VERSION;  # see L<perlmodstyle>

sub DEBUG () { 0 }

###########################################################
sub new
{
    my( $package ) = @_;

    return bless {}, $package;
}

###########################################################
sub init
{
    my( $self ) = @_;
    $self->init_args;
    $self->{cfg_dir} ||= $self->default_cfg_dir;
}

###########################################################
sub init_args
{
    my( $self ) = @_;
    $self->{VERBOSE} = 0;
    $self->{cfg_dir} = '';

    my $ok = GetOptions ( "verbose!" => \$self->{VERBOSE},
                          "cfg_dir=s" => \$self->{cfg_dir} );
    die "Error in options" unless $ok;
}

sub default_cfg_dir
{
    my( $self ) = @_;
    return File::Spec->catdir( File::HomeDir->my_home, qw( etc this2that.d ) );
}

###########################################################
sub run
{
    my( $self ) = @_;
    $self->build_inotify;
    $self->load_config;
    $self->inotify_config;
    while( $self->inotify->poll or $self->{loop} ) {
        $self->{loop} = 0;
    }
    $self->LOG( "Exiting" );
}

###########################################################
sub build_inotify
{
    my( $self ) = @_;
    $self->{inotify} = Linux::Inotify2->new;
}

sub inotify { $_[0]->{inotify} }

###########################################################
sub make_dir
{
    my( $self, $dir ) = @_;
    return if -d $dir;
    $self->doing( "Creating $dir" );
    make_path $dir;
}

###########################################################
sub load_config
{
    my( $self ) = @_;
    $self->make_dir( $self->{cfg_dir} );
    my $dir = $self->{cfg_dir};
    $self->LOG( "Reading $dir" );
    my $dh = IO::Dir->new( $dir ) or die "Unable to read $dir: $!";
    my $entry;
    while( defined( $entry = $dh->read ) ) {
        next if $entry =~ /^\./ or $entry =~ /(\.ba?k|~)$/;
        my $cfg = File::Spec->catfile( $dir, $entry );
        next unless -f $cfg;
        $self->config_setup( $cfg );
    }
}

###########################################################
sub inotify_config
{
    my( $self ) = @_;

    my $dir = $self->{cfg_dir};
    $self->LOG( "Watching $dir/" );
    $self->inotify->watch( $dir, IN_MOVED_TO|IN_CLOSE_WRITE, 
                                    sub { $self->cfg_dir_changed( @_ ) } );
}

###########################################################
sub config_setup
{
    my( $self, $file ) = @_;
    if( $self->have_config( $file ) ) {
        $self->config_remove( $file );
    }

    $self->doing( "Loading $file" );
    my $C = LoadFile $file;
    $self->config_init( $C );
    unless( $self->check_config( $file, $C ) ) {
        warn( "Config $file invalid\n" );
        return;
    }

    if( $C->{user} ) {
        my( $login, $pass, $uid, $gid ) = getpwnam( $C->{user} );
        $C->{uid} = $uid;
        $C->{gid} = $gid;
    }

    if( $C->{flag} ) {
        $C->{flag} = File::Spec->catfile( $C->{from}, $C->{flag} ) 
                        unless File::Spec->file_name_is_absolute( $C->{flag} );
    }

    $self->config_add( $file, $C );
    
    if( $C->{log} ) {
        my $dir = dirname $C->{log};
        $self->make_dir( $dir );
        if( $C->{user} ) {
            $self->__chown( $C, $dir );
            $self->__chown( $C, $C->{log} ) if -f $C->{log};
        }
    }
}

###########################################################
sub have_config
{
    my( $self, $file ) = @_;
    return exists $self->{file2C}{ $file };
}

###########################################################
sub config_init
{
    my( $self, $C ) = @_;
    if( $C->{constraint} ) {
        $self->constraint_init( $C );
    }
}

###########################################################
sub check_config
{
    my( $self, $file, $C ) = @_;
    my $ok = 1;
    unless( $C->{from} ) {
        warn "Missing from:";
        $ok = 0;
    }
    unless( $C->{to} or $C->{exec} ) {
        warn "Missing to: or exec:";
        $ok = 0;
    }
    if( $C->{to} and $C->{to} eq $C->{from} ) {
        warn "to: and from: are identical";
        $ok = 0;
    }
    if( $C->{delay} and $C->{delay} !~ /^\d+$/ ) {
        warn "delay: is not a number";
        $ok = 0;
    }
    if( $self->{dir2C}{ $C->{from} } ) {
        warn "Already watching $C->{from} from $self->{dir2C}{ $C->{from} }{cfg_file}";
        $ok = 0;
    }
    unless( $self->check_constraint( $C ) ) {
        warn "Bad constraint";
        $ok = 0;
    }

    return $ok;
}

###########################################################
sub config_add
{
    my( $self, $file, $C ) = @_;

    $C->{cfg_file} = $file;
    $self->{file2C}{ $file } = $C;

    # watch the config file for changes
    $self->LOG( "Watching config $file" );
    $C->{watch_cfg} = $self->inotify->watch( $file, 
                                            IN_MOVE_SELF|IN_CLOSE_WRITE|IN_DELETE_SELF, 
                                            sub { $self->config_changed( @_ ) } )
                        or die "Unable to watch config $file:  $!";

    my $dir = $C->{from};
    confess "Empty config ", pp $C unless $dir;
    $self->{dir2C}{$dir} = $C;

    # Create the source dir
    $self->make_dir( $C->{from} );

    # Create the dest dir
    if( $self->is_exec( $C ) ) {
        my $dir = $self->work_directory( $C );
        $self->make_dir( $dir );
        $self->__chown( $C, $dir );
    }
    else {
        $self->make_dir( $C->{to} );
        $self->__chown( $C, $C->{to} );
    }
 
    # do an initial check on the dir
    $self->hardcheck_dir( $dir );
    return if $C->{removed};        # some error happened

    # watch the dir
    if( $self->is_subdirs( $C ) ) {
        $self->LOG( "Watching $dir/" );
        $self->delay_add( $C );
        $C->{watch_dir} = $self->inotify->watch( $dir,
                                            IN_CLOSE_WRITE,
                                            sub { $self->delay_changed( $dir, @_ ) } )
                        or die "Unable to watch directory $dir: $!";  
    }
    else {
        # watch the directory for new stuff
        $self->LOG( "Watching $dir/" );
        my $sub;
        if( $C->{delay} ) {
            $sub = sub { $self->delay_dir_changed( $dir, @_ ) };
        }
        else {
            $sub = sub { $self->dir_changed( $dir, @_ ) };
        }

        $C->{watch_dir} = $self->inotify->watch( $dir, (IN_MOVED_TO|IN_CLOSE_WRITE|IN_DELETE), $sub )
                        or die "Unable to watch directory $dir: $!";  
    }
}


###########################################################
sub config_remove
{
    my( $self, $file ) = @_;
    $file = $file->{cfg_file} if ref $file;

    my $C = delete $self->{file2C}{ $file };

    # unwatch the config file for changes
    $C->{watch_cfg}->cancel if $C->{watch_cfg};
    delete $C->{watch_cfg};

    my $dir = $C->{from};
    delete $self->{dir2C}{ $dir };

    # unwatch the directory for new stuff
    $C->{watch_dir}->cancel if $C->{watch_dir};
    delete $C->{watch_dir};
    $C->{removed} = 1;
}


###########################################################
sub constraint_init
{
    my( $self, $C ) = @_;
    my $cstr = {};
    my $r = ref $C->{constraint};
    if( 'HASH' eq $r ) {
        $cstr = $C->{constraint};
    }
    elsif( 'ARRAY' eq $r ) {
        foreach my $e ( @{ $C->{contraint} } ) {
            my $t = ref $e;
            if( 'HASH' eq $t ) {
                @{ $cstr }{ keys %$t } = values %$t;
            }
            else {
                $cstr->{$e} = 1;
            }
        }
    }
    elsif( defined $C->{constraint} ) {
        $cstr->{ $C->{constraint} } = 1;
    }

    $C->{constraint} = $cstr;

    if( $cstr->{'required-ext'} ) {
        my $ext = $cstr->{'required-ext'};
        $r = ref $ext;
        unless( $r ) {
            $ext = { $ext=>1 };
        }
        elsif( 'ARRAY' eq $r ) {
            my %new;
            @new{ @$ext } = (1) x @$ext;
            $ext = \%new;
        }
        $cstr->{'required-ext'} = $ext;
        $cstr->{'required-ext-re'} = join '|', map quotemeta, map { s/^\.+//; $_ } keys %$ext;
        $cstr->{'group-filename'} = 1;
    }
    else {
        delete $cstr->{'required-ext-re'};
    }
}

###########################################################
# constraint:
#       max-dest: 100
#       retry-delay: XX
#       src-subdirs: 1
#       group-filename: 1
#       numeric-sort: 1
my %ALLOWED_CONSTRAINTS = ( 
        'max-dest'      => 1, 
        'retry-delay'   => 1, 
        'src-subdirs'   => 1, 
        'group-filename' => 1,
        'numeric-sort'  => 1,
        'mtime-sort'    => 1,
        'sort'          => 1,
        'reverse'       => 1,
        'required-ext'  => 1,
        'required-ext-re'  => 1
    );
sub check_constraint
{
    my( $self, $C ) = @_;
    return 1 unless $C->{constraint};
    my $ok = 1;
    foreach my $c ( keys %{ $C->{constraint} } ) {
        next if $ALLOWED_CONSTRAINTS{ $c };
        warn "Unknown constraint: $c";
        $ok = 0;
    }

    my $cstr = $C->{constraint};
    foreach my $k ( qw( retry-delay max-dest ) ) {
        next unless $cstr->{$k};
        next if looks_like_number $cstr->{$k};
        warn "constraint/$k isn't a number";
        $ok = 0;
    }

    return $ok;
}

###########################################################
sub is_ok_constraint
{
    my( $self, $C ) = @_;
    my $cstr = $C->{constraint};
    if( $cstr->{'max-dest'} ) {
        return 0 if $self->file_count( $C->{to} ) > $cstr->{'max-dest'};
    }
    return 1;
}

sub is_max_dest
{
    my( $self, $C ) = @_;
    return !! $C->{constraint}{'max-dest'};
}

sub is_group_filename
{
    my( $self, $C ) = @_;
    return !! $C->{constraint}{'group-filename'};
}

sub is_subdirs
{
    my( $self, $C ) = @_;
    return !! $C->{constraint}{'src-subdirs'};
}

sub is_ext_ok
{
    my( $self, $C, $list ) = @_;
    my $re = $C->{constraint}{'required-ext-re'};
    return 1 unless $re;

    my %need = %{ $C->{constraint}{'required-ext'} };
    foreach my $file ( @$list ) {
        my $ext = '';
        $ext = $1 if $file =~ /\.($re)$/;
        delete $need{ $ext };
    }
    return 1 unless keys %need;
    $self->LOG( "Missing: ", join ', ', sort keys %need );
    return;
}

###########################################################
sub is_exec
{
    my( $self, $C ) = @_;
    return !! $C->{exec};
}



###########################################################
sub hardcheck_dir
{
    my( $self, $dir ) = @_;
    $self->doing( "Check $dir" );
    my $C = $self->{dir2C}{ $dir };
    return if $self->flag_file( $C );
    return unless $self->is_ok_constraint( $C );

    if( $self->is_subdirs( $C ) ) {
        $self->doing( $dir );
        $self->check_subdirs( $dir );
    }
    else {
        $self->handle_dir( $C, $dir );
    }
}

###########################################################
sub check_subdirs
{
    my( $self, $dir ) = @_;
    my $C = $self->{dir2C}{ $dir };

    my @todo = $self->read_dir( $C, $dir, sub { -d $_[1] } );

    foreach my $full ( @todo ) {
        last if $self->handle_dir( $C, $full );
    }
}    

###########################################################
# return true if max files were copied
sub handle_dir
{
    my( $self, $C, $dir ) = @_;

    # max-dest: 100
    my $max;
    my $count;
    if( $self->is_max_dest( $C ) ) {
        $count = $self->file_count( $C->{to} );
        $max = $C->{constraint}{'max-dest'};
        return 1 if $count > $max;
    }

    # contraint/group-filenames
    my $is_group = $self->is_group_filename( $C );
    # Note that require-ext will always turn on group-filenames

    my %todo;
    my @keys;
    foreach my $file ( $self->read_dir( $C, $dir, sub { -f $_[1] } ) ) {
        return 1 if defined $max and $count > $max;
        if( $is_group ) {
            my $entry = basename $file;
            $entry =~ /^([^.]+)/;
            my $basename = $1||$entry;
            unless( $todo{$basename} ) {
                push @keys, $basename;
                $todo{$basename} = [$file];
                # each entry in %todo has a minimum of 1 file, so we can exit early
                last if defined $max and 0+@keys > $max;
            }
            else {
                push @{ $todo{$basename} }, $file;
            }
        }
        else {
            $self->file_handle( $C, $file );
            $count++;
            return if $C->{removed};
        }
    }
    return unless $is_group;

    foreach my $base ( @keys ) {
        return if $C->{removed};

        return 1 if defined $max and $count > $max;
        next unless $self->is_ext_ok( $C, $todo{$base} );
    
        if( $self->is_exec( $C ) ) {
            $self->file_exec( $C, $todo{$base} );
        }
        else {
            foreach my $file ( @{ $todo{$base} } ) {
                $self->file_handle( $C, $file );
                $count++;
            }
        }
    }
    return;
}

###########################################################
# Read all the entries to a directory, with filtering
# Sort the entries according to constraints
sub read_dir
{
    my( $self, $C, $dir, $filter ) = @_;

    my $dh = IO::Dir->new( $dir ) or die "Unable to read $dir: $!";
    my @list;
    my $entry;
    my $is_exec = 1 if $self->is_exec( $C );

    # read all the entries
    while( defined( $entry = $dh->read ) ) {
        next if $entry =~ /^\./;
        next if $is_exec and $entry eq 'WORK';  # skip working directory

        my $full = File::Spec->catfile( $dir, $entry );
        next if $filter and not $filter->( $C, $full, $entry );

                    #  0      1
        push @list, [ $full, $entry ];
        if( $C->{constraint}{'numeric-sort'} ) {
            $list[-1][3] = 1 if $entry =~ /^(\d+)/;
            $list[-1][4] = $1;
        }
        elsif( $C->{constraint}{'time-sort'} ) {
            $list[-1][2] = (stat $full)[9];
        }
    }

    # (possibly) sort the entries
    if( $C->{constraint}{'numeric-sort'} ) {
        @list = sort { $a->[3] ? ( $b->[3] ? (($a->[4] <=> $b->[4]) or ($a->[1] cmp $b->[1])) : -1 ) 
                               : ( $b->[3] ? 1 : $a->[1] cmp $b->[1] ) } @list;
    }
    elsif( $C->{constraint}{'sort'} ) {
        @list = sort { $a->[1] cmp $b->[1] } @list;
    }
    elsif( $C->{constraint}{'time-sort'} ) {
        @list = sort { $a->[2] <=> $b->[2] } @list;
    }
    if( $C->{constraint}{'reverse'} ) {
        @list = reverse @list;
    }

    
    return map { $_->[0] } @list;
}

###########################################################
sub config_changed
{
    my( $self, $e ) = @_;
    my $file = $e->fullname;
    if( $e->IN_MOVE_SELF ) {
        $self->doing( "(config) Moved $file" );
        $self->config_remove( $file );
    }
    if( $e->IN_CLOSE_WRITE ) {
        $self->doing( "(config) Closed $file" );
        $self->config_setup( $file );
    }
    if( $e->IN_DELETE_SELF ) {
        $self->doing( "(config) Delete $file" );
        $self->config_remove( $file );
    }

    return;
    if( -f $file ) {
        $self->doing( "Config $file changed" );
        $self->config_setup( $file );
    }
    else {
        $self->doing( "Config $file move" );
        $self->config_remove( $file );
    }
}

###########################################################
sub alarm_add
{
    my( $self, $C, $delay, $sub ) = @_;

    $C->{alarm} = 0+$sub;
    setalarm( $delay, $sub );
    $self->LOG( "Created an alarm for dir=$C->{from} sub=$C->{alarm}" );
}

sub alarm_remove
{
    my( $self, $C ) = @_;
    my $want = delete $C->{alarm};
    return unless $want;
    for( my $N = 0; 1; $N++ ) {
        my $H = Alarm::Concurrent::gethandler( $N );
        unless( defined $H ) {
            $self->LOG( "Can't find alarm for dir=$C->{from} sub=$want" );
        }
        if( $want == 0+$H ) {
            $self->LOG( "Clear alarm N=$N" );
            Alarm::Concurrent::clearalarm( $N );
            return;
        }
        $N++;
    }
}

###########################################################
sub delay_dir_changed
{
    my( $self, $dir, $e ) = @_;
    my $C = $self->{dir2C}{ $dir };
    return unless $C;
    my $delay = $C->{delay} || 5;
    # remove any previously set alarm
    if( $C->{alarm} ) {
        $self->alarm_remove( $C ) 
    }
    else {
        $self->doing( "Inotify $dir" );
    }
    # create an alarm
    $self->alarm_add( $C, $delay, sub { $self->delay_dir_alarm( $dir, $e ) } );
}

sub delay_dir_alarm
{
    my( $self, $dir, $e ) = @_;
    DEBUG and $self->LOG( "Alarm for $dir fullfile=" . $e->fullname );
    $self->{loop} = 1;                  # don't exit ->pool
    my $C = $self->{dir2C}{ $dir };
    return unless $C;
    delete $C->{alarm};                 # forget we ever had an alarm
    $self->dir_changed( $dir, $e );     # now we finally do the work
}

###########################################################
sub dir_changed
{
    my( $self, $dir, $e ) = @_;
    my $C = $self->{dir2C}{ $dir };
    return unless $C;
    if( $e->IN_DELETE ) {
        DEBUG and $self->LOG( "delete" );
        if( $self->is_flag_gone( $e, $C ) ) { 
            # We used to have a flag but not any more
            $self->doing( "Flag $C->{flag} disapeared" );
            $self->hardcheck_dir( $dir );
        }
    }
    elsif( not $self->is_ok_constraint( $C ) ) {
        return;
    }
    elsif( $C->{delay} or $self->is_group_filename( $C ) ) {
        DEBUG and $self->doing( "$dir has changed" );
        return $self->hardcheck_dir( $dir );
    }
    else {
        my $file = $e->fullname;
        DEBUG and $self->doing( "$file ($dir) is new" );
        $self->file_handle( $C, $file );
    }
}

###########################################################
sub cfg_dir_changed
{
    my( $self, $e ) = @_;
    my $file = $e->fullname;
    return unless -f $file;

    if( $e->IN_MOVED_FROM ) {
          $self->doing( "(dir) From $file" );
    }
    if( $e->IN_MOVED_TO ) {
        $self->doing( "(dir) To $file");
        $self->doing( "Config $file is new");
        $self->config_setup( $file );
    }
    if( $e->IN_CLOSE_WRITE ) {
        DEBUG and $self->doing( "(dir) New $file");
        DEBUG and $self->doing( "Config $file is new");
        $self->config_setup( $file );
    }
    if( $e->IN_CREATE ) {
        $self->doing( "(dir) Create $file");
    }
    if( $e->IN_DELETE ) {
        $self->doing( "Delete $file");
    }

    return;
}

###########################################################
# We are waiting for 'retry-delay' to see if subdirectories were created
# Only used if we have src-subdirs: 1

# Returns a filename to use as a flag
sub delay_flag
{
    my( $self, $C ) = @_;
    return File::Spec->catfile( $C->{from}, join '-', ref( $self), "ALARMFLAG" );
}

# Create the delay
sub delay_add
{
    my( $self, $C ) = @_;

    my $flag = $self->delay_flag( $C );
    unlink $flag;    

    my $delay = $C->{constraint}{'retry-delay'} || 60;
    $self->LOG( "Retry in $delay seconds" );
    setalarm( $delay, sub { $self->delay_handler( $C ) } );
}

# This is ALARM handler.  It creates a flag that will trigger ->delay_changed
sub delay_handler
{
    my( $self, $C ) = @_;
    return if $C->{removed};
    $self->LOG( "ALARM for $C->{from}/" );

    my $file = $self->delay_flag( $C );
    my $fh = IO::File->new;
    $fh->open( $file, ">" ) or die "Unable to create $file: $!";
    $fh->print( time );
    $self->{loop} = 1; # the alarm is killing inotify->poll    

    return;
}

# Triggered by the file created in ->delay_handler
sub delay_changed
{
    my( $self, $dir, $e ) = @_;
    my $file = $e->fullname;
    my $C = $self->{dir2C}{ $dir };
    my $flag = $self->delay_flag( $C );

    if( $flag eq $file ) {      
        $self->LOG( "Retry $dir/" );
        unlink $flag;    
        $self->hardcheck_dir( $dir );
        $self->delay_add( $C );
    }
    # ignore other files
}



###########################################################
# Make sure the flag file is missing.
# If it's present, then remember it and do nothing
sub flag_file
{
    my( $self, $C ) = @_;
    return unless $C->{flag};
    if( -f $C->{flag} ) {
        $self->doing( "Flag $C->{flag} exists") unless $self->{had_flag};
        $C->{had_flag} = 1;
        return 1;
    }
    return 0;
}

# Return true if the flag file is now gone
sub is_flag_gone
{
    my( $self, $e, $C ) = @_;

    return unless $C->{flag};
    if( $C->{had_flag} ) {     # flag file existed?
        DEBUG and $self->LOG( "we had_flag" );
        return if -f $C->{flag};                        # flag file still exists?
        delete $C->{had_flag};                          # it doesn't
        return 1;
    }
    elsif( $e->fullname eq $C->{flag} ) {
        DEBUG and $self->LOG( "Is $C->{flag} gone?" );
        return !-f $C->{flag};
    }
    $self->LOG( $e->fullname, " and we wanted $C->{flag}" );
    return;
}



###########################################################
sub file_count
{
    my( $self, $dir ) = @_;
    my $dh = IO::Dir->new( $dir ) or die "Unable to open $dir: $!";
    my $entry;
    my $count = 0;
    while( defined( $entry = $dh->read ) ) {
        next if $entry =~ /^\./;
        my $full = File::Spec->catfile( $dir, $entry );
        $count++ if -f $full;
    }
    return $count;
}

###########################################################
# Process one file.  Either with exec: or moving to to:
sub file_handle
{
    my( $self, $C, $file ) = @_;
    # Don't process if the flag file is active
    return if $self->flag_file( $C );
    if( $self->is_exec( $C ) ) {
        $self->file_exec( $C, [$file] );
    }
    else {
        $self->file_move( $C, $file );
    }
}

###########################################################
# Move a file to to:
sub file_move
{
    my( $self, $C, $file ) = @_;
    my $basename = basename $file;
    my $to = File::Spec->catfile( $C->{to}, $basename );
    if( -f $to ) {
        $C->{quiet} or $self->doing( "Unlink $to" );
        unlink $to or $self->doing( "Unable to unlink $to: $!");
    }
    $C->{quiet} or $self->doing( "$file -> $to" );
    if( move $file, $to ) {
        $self->__chown( $C, $to );
        return $to; 
    }
    else {
        $self->doing( "Failure: $!" );
        return;
    }
}


sub __chown
{
    my( $self, $C, $to ) = @_;
    return unless $C->{uid} or $C->{gid};

    my $uid = $C->{uid} || -1;
    my $gid = $C->{gid} || -1;

    my @stat = stat $to;
    return if $stat[4] == $uid and $stat[5] == $gid;    
    
    $C->{quiet} or $self->doing( "chown $uid:$gid $to" );
    
    chown $uid, $gid, $to or $self->doing( "Unable to chown $to: $!");
}


###########################################################

# Return a directory for temporary work files
sub work_directory
{
    my( $self, $C ) = @_;
    return File::Spec->catdir( $C->{from}, 'WORK' );
}

# Execute a command with a file
sub file_exec
{
    my( $self, $C, $list ) = @_;

    # Copy to a WORK dir
    my @working;
    {
        local $C->{quiet} = !$self->{VERBOSE};
        # this will fail with constraint/subdirs 
        local $C->{to} = $self->work_directory( $C );
        foreach my $file ( @$list ) {
            push @working, $self->file_move( $C, $file );
        }
        unless( @working == @$list ) {
            $self->doing( "Failed to move all files to working directory $C->{to}" );
            return;
        }
    }

    my $exec = $C->{exec};
    my $files = join ' ', map {shell_quote $_} sort @working;

    unless( $exec =~ s/\[file\]/$files/ ) {
        $exec .= " $files";
    }

    if( $C->{log} ) {
        $exec .= " >>$C->{log}";
    }

    if( $C->{user} ) {
        $exec = qq(su - $C->{user} -c ) . shell_quote $exec;
    }

    $self->doing( $exec );
    my $rv = system $exec;
    if( $rv == 0 ) {
        foreach my $file ( @working ) {
            next unless -f $file;
            $self->LOG( "Unlink $file" );
            unlink $file;
        }
    }
    else {
        $self->LOG( sprintf "Error: 0x%X", $rv );
        $self->LOG( "Droping config $C->{cfg_file}" );
        $self->config_remove( $C );
    }
}


###########################################################
sub doing
{
    my $self = shift;
    my $text = join " ", @_;
    $text =~ s/\n+$//;
    print strftime "%Y/%m/%d %H:%M:%S $text\n", localtime;
}

###########################################################
sub LOG
{
    my $self = shift;
    return unless $self->{VERBOSE};
    $self->doing( @_ );
}

1;

__END__
# Below is stub documentation for your module. You'd better edit it!

=head1 NAME

App::This2that - Perl extension for blah blah blah

=head1 SYNOPSIS

  use App::This2that;
  blah blah blah

=head1 DESCRIPTION

Stub documentation for App::This2that, created by h2xs. It looks like the
author of the extension was negligent enough to leave the stub
unedited.

Blah blah blah.

=head2 EXPORT

None by default.



=head1 SEE ALSO

Mention other useful documentation such as the documentation of
related modules or operating system documentation (such as man pages
in UNIX), or any relevant external documentation such as RFCs or
standards.

If you have a mailing list set up for your module, mention it here.

If you have a web site set up for your module, mention it here.

=head1 AUTHOR

Philip Gwyn, E<lt>fil@localdomainE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2015 by Philip Gwyn

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.10.1 or,
at your option, any later version of Perl 5 you may have available.


=cut
